from torch import nn, optim
import math
import torch
import torch.nn.functional as F


class SirenNet(nn.Module):
    """Sinusoidal Representation Network (SIREN) with residual connections (ReSIREN).
    Adapted from: https://github.com/microsoft/satclip/blob/main/satclip/location_encoder.py

        residual_connections: Turn the network into a ReSIREN model
        h_siren: Use the H-SIREN activation after the first layer
        return_hidden_embs: Append lower-resolution embeddings from lower layers to the output
    """

    def __init__(
        self,
        dim_in,
        dim_hidden,
        dim_out,
        num_layers,
        w0=1.0,
        w0_initial=30.0,
        use_bias=True,
        final_activation=None,
        dropout=False,
        residual_connections=False,
        h_siren=False,
        return_hidden_embs=None,
    ):
        super().__init__()
        self.num_layers = num_layers
        self.dim_hidden = dim_hidden
        self.return_hidden_embs = return_hidden_embs

        self.layers = nn.ModuleList([])
        for ind in range(num_layers):
            is_first = ind == 0
            layer_w0 = w0_initial if is_first else w0
            layer_dim_in = dim_in if is_first else dim_hidden

            self.layers.append(
                Siren(
                    dim_in=layer_dim_in,
                    dim_out=dim_hidden,
                    w0=layer_w0,
                    use_bias=use_bias,
                    is_first=is_first,
                    h_siren=h_siren,
                    dropout=dropout,
                    residual_connections=residual_connections,
                )
            )

        final_activation = nn.Identity() if not final_activation else final_activation
        self.last_layer = Siren(
            dim_in=dim_hidden,
            dim_out=dim_out,
            w0=w0,
            use_bias=use_bias,
            activation=final_activation,
            dropout=False,
        )

    def forward(self, x, mods=None):

        res = []
        gaussian = None
        # Passing the gaussian (vector after dot product and before sin) between layers as residual connection
        for i in range(len(self.layers)):
            layer = self.layers[i]
            x, gaussian = layer(x, gaussian)
            if self.return_hidden_embs is not None:
                if i in self.return_hidden_embs:
                    res.append(x)

        x, _ = self.last_layer(x, gaussian)

        if len(res) > 0:
            res.append(x)
            x = torch.cat(res, dim=1)
        return x


class Sine(nn.Module):
    def __init__(self, w0=1.0):
        super().__init__()
        self.w0 = w0

    def forward(self, x):
        return torch.sin(self.w0 * x)


class Siren(nn.Module):
    def __init__(
        self,
        dim_in,
        dim_out,
        w0=1.0,
        c=6.0,
        is_first=False,
        use_bias=True,
        activation=None,
        dropout=False,
        residual_connections=False,
        h_siren=False,
    ):
        super().__init__()
        self.dim_in = dim_in
        self.is_first = is_first
        self.dim_out = dim_out
        self.dropout = dropout
        self.h_siren = h_siren

        weight = torch.zeros(dim_out, dim_in)
        bias = torch.zeros(dim_out) if use_bias else None
        self.init_(weight, bias, c=c, w0=w0)

        self.weight = nn.Parameter(weight)
        self.bias = nn.Parameter(bias) if use_bias else None
        self.activation = Sine(w0) if activation is None else activation

        self.residual_connections = residual_connections

    def init_(self, weight, bias, c, w0):
        dim = self.dim_in

        w_std = (1 / dim) if self.is_first else (math.sqrt(c / dim) / w0)
        weight.uniform_(-w_std, w_std)

        if not bias is None:
            bias.uniform_(-w_std, w_std)

    def forward(self, x, prev_gaussian=None):
        # Apply base linear transform
        out = F.linear(x, self.weight, self.bias)
        # Use dropout (not recommended for reconstruction)
        if self.dropout:
            out = F.dropout(out, training=self.training)
        # If using ReSIREN, merge with the previous layers gaussian
        if (
            self.residual_connections
            and prev_gaussian is not None
            and out.shape == prev_gaussian.shape
        ):
            out = (out + prev_gaussian) / 2
        # Save the gaussian to pass it on
        gaussian = out
        # Apply the additional transformation for H-SIREN
        if self.h_siren and self.is_first:
            out = torch.sinh(2 * out)
        # Apply the activation (for nearly all layers Sine)
        out = self.activation(out)
        return out, gaussian



from huggingface_hub import get_collection
import ssl

#HACK: Ignore SSL errors
ssl._create_default_https_context = ssl._create_unverified_context


def list_models(model_name):
    collection = get_collection("MVRL/remote-sensing-foundation-models-664e8fcd67d8ca8c03f42d00")
    models = filter(lambda item: model_name.lower() in item.item_id.lower(), collection.items)
    print(f"Available {model_name} pretrained models:\n")
    for model_info in models:
        print(model_info.item_id)

def help(model):
    print(model.__doc__)


import torch

from .direct import Direct
from .siren import SirenNet
from transformers import PretrainedConfig
from huggingface_hub import PyTorchModelHubMixin

VAR_NAMES = [
    "cmi",
    "clt",
    "hurs",
    "pet",
    "pr",
    "rsds",
    "sfcWind",
    "tas",
    "tasmax",
    "tasmin",
    "vpd",
]

CHELSA_MEAN = torch.tensor(
    [
        [-264.1493656],
        [3912.44628016],
        [5921.65964573],
        [9385.47468266],
        [697.03653109],
        [15219.37926928],
        [3498.8511804],
        [2819.56006368],
        [2864.08583811],
        [2773.46759638],
        [8039.37322797],
    ]
).T

CHELSA_STD = torch.tensor(
    [
        [1042.67560332],
        [1767.94018571],
        [1185.91587823],
        [6639.79069994],
        [883.56243405],
        [7843.49167037],
        [1637.09237995],
        [174.43791946],
        [181.69448751],
        [167.07485901],
        [7516.98198719],
    ]
).T


class Climplicit(torch.nn.Module, PyTorchModelHubMixin):
    """
    CLIMPLICIT IMPLICITLY ENCODES GLOBAL CLIMATIC INFORMATION.

    config: Dict
    -> "return_chelsa": bool, defines whether the implicit embeddings
        or original CHELSA reconstructions should be returned by the forward pass.

    EXAMPLE USAGE:

    import torch
    from rshf.climplicit import Climplicit

    model = Climplicit.from_pretrained("Jobedo/climplicit", config={"return_chelsa": False})

    loc = [8.550155, 47.396702]  # Lon/Lat or our office
    april = 4  # April
    batchsize = 10

    # Call with a month
    month = torch.ones(batchsize) * april
    print("Output shape with month:", model(torch.tensor([loc] * batchsize), month).shape)
    # >>> Output shape with month: torch.Size([10, 256])

    # Call without month
    print("Output shape without month:", model(torch.tensor([loc] * batchsize)).shape)
    # >>> Output shape without month: torch.Size([10, 1024])

    # Return the CHELSA reconstruction instead of Climplicit embeddings
    model = Climplicit.from_pretrained("Jobedo/climplicit", config={"return_chelsa": True})
    print("Output shape of CHELSA reconstruction with month:", model(torch.tensor([loc] * batchsize), month).shape)
    # >>> Output shape of CHELSA reconstruction with month: torch.Size([10, 11])
    """
    def __init__(self, config: PretrainedConfig):
        super().__init__()
        self.config = config
        if type(config) is dict:
            self.config = PretrainedConfig().from_dict(config)

        self.location_encoder = SirenNet(
            dim_in=4,
            dim_hidden=512,
            dim_out=256,
            num_layers=16,
            dropout=False,
            h_siren=True,
            residual_connections=True,
        )
        
        self.pos_embedding = Direct(lon_min=-180, lon_max=180, lat_min=-90, lat_max=90)

        self.chelsa_regressor = torch.nn.Linear(256, 11)        

        self.return_chelsa = self.config.return_chelsa

    def forward(self, coordinates, month=None):
        # Apply the positional embedding
        loc = self.pos_embedding(coordinates)

        if month is None:
            res = []
            # Get the Climplicit embeddings for four months across the year
            for m in [3, 6, 9, 12]:
                month = torch.ones(len(coordinates)) * m
                loc_month = torch.concat(
                    [
                        loc,
                        torch.sin(month / 12 * torch.pi * 2).unsqueeze(dim=-1).to(loc.device),
                        torch.cos(month / 12 * torch.pi * 2).unsqueeze(dim=-1).to(loc.device),
                    ],
                    dim=-1,
                )
                x = self.location_encoder(loc_month)
                if self.return_chelsa:
                    x = self.chelsa_regressor(x)
                    x = x * CHELSA_STD + CHELSA_MEAN
                res.append(x)
            return torch.cat(res, dim=-1)

        # If we have a month
        # Append the month to the positional embedding
        loc_month = torch.concat(
            [
                loc,
                torch.sin(month / 12 * torch.pi * 2).unsqueeze(dim=-1),
                torch.cos(month / 12 * torch.pi * 2).unsqueeze(dim=-1),
            ],
            dim=-1,
        )
        # Return the Climplicit embedding
        x = self.location_encoder(loc_month)
        if self.return_chelsa:
            x = self.chelsa_regressor(x)
            x = x * CHELSA_STD + CHELSA_MEAN
        return x


if __name__ == "__main__":
    print(Climplicit.from_pretrained("Jobedo/climplicit", config={"return_chelsa": False}))



import torch
import torch.nn as nn


class Direct(nn.Module):
    """
    Just turn it into a [-1,1] scaling based on the input range.
    """

    def __init__(self, lon_min, lon_max, lat_min, lat_max):
        """
        Args:
        """
        super(Direct, self).__init__()
        self.lon_min = lon_min
        self.lon_max = lon_max
        self.lat_min = lat_min
        self.lat_max = lat_max

    def forward(self, coords):
        """"""
        lon, lat = coords[:, 0], coords[:, 1]
        lon = 2 * (lon - self.lon_min) / (self.lon_max - self.lon_min) - 1
        lat = 2 * (lat - self.lat_min) / (self.lat_max - self.lat_min) - 1
        return torch.stack([lon, lat], dim=1).float()



import torch.nn as nn
import torch
from huggingface_hub import PyTorchModelHubMixin
from transformers import AutoModel, PretrainedConfig
from torchvision import transforms

class Dinov3_Sat(nn.Module, PyTorchModelHubMixin):
    def __init__(self, config: PretrainedConfig):
        super().__init__()
        self.config = config
        if type(config) is dict:
            self.config = PretrainedConfig().from_dict(config)
        self.model = AutoModel.from_pretrained(self.config.model_name)
    
    def transform(self, x, size=224):
        preprocess = transforms.Compose([
            transforms.Resize(size),
            transforms.CenterCrop(size),
            transforms.Normalize(mean=self.config.img_mean, std=self.config.img_std),
        ])
        return preprocess(x)
        
    
    def forward(self, x):
        return self.model(x)